// Python Tools for Visual Studio
// Copyright(c) Microsoft Corporation
// All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the License); you may not use
// this file except in compliance with the License. You may obtain a copy of the
// License at http://www.apache.org/licenses/LICENSE-2.0
//
// THIS CODE IS PROVIDED ON AN  *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS
// OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY
// IMPLIED WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
// MERCHANTABLITY OR NON-INFRINGEMENT.
//
// See the Apache Version 2.0 License for specific language governing
// permissions and limitations under the License.

using System;
using System.Collections.Generic;
using Microsoft.PythonTools.Editor.Core;
using Microsoft.VisualStudio.ComponentModelHost;
using Microsoft.VisualStudio.Language.StandardClassification;
using Microsoft.VisualStudio.Text;
using Microsoft.VisualStudio.Text.Adornments;
using Microsoft.VisualStudio.Text.Editor;
using Microsoft.VisualStudio.Text.Tagging;

namespace Microsoft.PythonTools.Editor {
    /// <summary>
    /// Provides highlighting of matching braces in a text view.
    /// </summary>
    class BraceMatcher {
        private readonly ITextView _textView;
        private readonly IComponentModel _compModel;
        private ITextBuffer _markedBuffer;
        private static TextMarkerTag _tag = new TextMarkerTag("Brace Matching (Rectangle)");

        /// <summary>
        /// Starts watching the provided text view for brace matching.  When new braces are inserted
        /// in the text or when the cursor moves to a brace the matching braces are highlighted.
        /// </summary>
        public static void WatchBraceHighlights(ITextView view, IComponentModel componentModel) {
            var matcher = new BraceMatcher(view, componentModel);

            // position changed only fires when the caret is explicitly moved, not from normal text edits,
            // so we track both changes and position changed.
            view.Caret.PositionChanged += matcher.CaretPositionChanged;
            view.TextBuffer.Changed += matcher.TextBufferChanged;
            view.Closed += matcher.TextViewClosed;
        }

        public BraceMatcher(ITextView view, IComponentModel componentModel) {
            _textView = view;
            _compModel = componentModel;
        }

        private void TextViewClosed(object sender, EventArgs e) {
            _textView.Caret.PositionChanged -= CaretPositionChanged;
            _textView.TextBuffer.Changed -= TextBufferChanged;
            _textView.Closed -= TextViewClosed;
        }

        private void CaretPositionChanged(object sender, CaretPositionChangedEventArgs e) {
            RemoveExistingHighlights();

            UpdateBraceMatching(e.NewPosition.BufferPosition.Position);
        }

        private void TextBufferChanged(object sender, TextContentChangedEventArgs changed) {
            RemoveExistingHighlights();

            if (changed.Changes.Count == 1) {
                var newText = changed.Changes[0].NewText;
                if (newText == ")" || newText == "}" || newText == "]") {
                    UpdateBraceMatching(changed.Changes[0].NewPosition + 1);
                }
            }
        }

        private bool HasTags {
            get {
                return _markedBuffer != null;
            }
        }

        private void RemoveExistingHighlights() {
            if (HasTags) {
                RemoveExistingHighlights(_markedBuffer);
                _markedBuffer = null;
            }
        }

        private void UpdateBraceMatching(int pos) {
            if (pos != 0) {
                var prevCharText = _textView.TextBuffer.CurrentSnapshot.GetText(pos - 1, 1);
                if (prevCharText == ")" || prevCharText == "]" || prevCharText == "}") {
                    if (HighlightBrace(GetBraceKind(prevCharText), pos, -1)) {
                        return;
                    }
                }
            }

            if (pos != _textView.TextBuffer.CurrentSnapshot.Length) {
                var nextCharText = _textView.TextBuffer.CurrentSnapshot.GetText(pos, 1);
                if (nextCharText == "(" || nextCharText == "[" || nextCharText == "{") {
                    HighlightBrace(GetBraceKind(nextCharText), pos + 1, 1);
                }
            }
        }

        private void RemoveExistingHighlights(ITextBuffer buffer) {
            if (HasTags) {
                GetTextMarker(buffer).RemoveTagSpans(x => true);
            }
        }

        private SimpleTagger<TextMarkerTag> GetTextMarker(ITextBuffer buffer) {
            return _compModel.GetService<ITextMarkerProviderFactory>().GetTextMarkerTagger(buffer);
        }

        private bool HighlightBrace(BraceKind brace, int position, int direction) {
            var pt = _textView.BufferGraph.MapDownToInsertionPoint(
                new SnapshotPoint(_textView.TextBuffer.CurrentSnapshot, position),
                PointTrackingMode.Positive,
                EditorExtensions.IsPythonContent
            );

            if (pt == null) {
                return false;
            }

            return HighlightBrace(brace, pt.Value, direction);

        }

        private bool HighlightBrace(BraceKind brace, SnapshotPoint position, int direction) {
            var classifier = position.Snapshot.TextBuffer.GetPythonClassifier();
            
            if (classifier != null) {
                var snapshot = position.Snapshot;
                var span = new SnapshotSpan(snapshot, position.Position - 1, 1);
                var originalSpan = span;

                var spans = classifier.GetClassificationSpans(span);
                // we don't highlight braces if we're in a comment or string literal
                if (spans.Count == 0 ||
                    (spans.Count == 1 &&
                    (!spans[0].ClassificationType.IsOfType(PredefinedClassificationTypeNames.String) &&
                    !spans[0].ClassificationType.IsOfType(PredefinedClassificationTypeNames.Comment)))) {

                    // find the opening span
                    var curLine = snapshot.GetLineFromPosition(position);
                    int curLineNo = curLine.LineNumber;

                    if (direction == 1) {
                        span = new SnapshotSpan(snapshot, position, curLine.End.Position - position);
                    } else {
                        span = new SnapshotSpan(curLine.Start, position - 1);
                    }

                    int depth = 1;
                    for (; ; ) {
                        spans = classifier.GetClassificationSpans(span);
                        for (int i = direction == -1 ? spans.Count - 1 : 0; i >= 0 && i < spans.Count; i += direction) {
                            if (spans[i].IsCloseGrouping()) {
                                if (IsSameBraceKind(spans[i].Span.GetText(), brace)) {
                                    depth -= direction;
                                }
                            } else if (spans[i].IsOpenGrouping()) {
                                if (IsSameBraceKind(spans[i].Span.GetText(), brace)) {
                                    depth += direction;
                                }
                            }

                            if (depth == 0) {
                                RemoveExistingHighlights();
                                _markedBuffer = snapshot.TextBuffer;

                                // left brace
                                GetTextMarker(snapshot.TextBuffer).CreateTagSpan(snapshot.CreateTrackingSpan(spans[i].Span, SpanTrackingMode.EdgeExclusive), _tag);
                                // right brace
                                GetTextMarker(snapshot.TextBuffer).CreateTagSpan(snapshot.CreateTrackingSpan(new SnapshotSpan(snapshot, position - 1, 1), SpanTrackingMode.EdgeExclusive), _tag);
                                return true;
                            }
                        }

                        curLineNo += direction;
                        if (curLineNo < 0 || curLineNo >= snapshot.LineCount) {
                            break;
                        }

                        var line = snapshot.GetLineFromLineNumber(curLineNo);
                        span = new SnapshotSpan(line.Start, line.End);
                    }
                }
            }
            return false;
        }

        private enum BraceKind {
            Bracket,
            Paren,
            Brace
        }

        private static bool IsSameBraceKind(string brace, BraceKind kind) {
            return GetBraceKind(brace) == kind;
        }

        private static BraceKind GetBraceKind(string brace) {
            switch (brace[0]) {
                case '[':
                case ']': return BraceKind.Bracket;
                case '(':
                case ')': return BraceKind.Paren;
                case '{':
                case '}': return BraceKind.Bracket;
                default: throw new InvalidOperationException();
            }
        }
    }
}
